/*
 * $Id: bgp_pdu.c,v 1.7 1998/05/29 19:57:27 masaki Exp $
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/uio.h>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <stdarg.h>
#include <netdb.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <memory.h>
#ifndef __linux__
#include <sys/conf.h>		/* BSDism */
#endif
#include <errno.h>

#include <mrt.h>
#include <timer.h>
#include <select.h>
#include <trace.h>
#include <io.h>
#include <bgp.h>


/*
 * return 0 if an error can be ignored 
 */

int
bgp_read (bgp_peer_t *peer, int sockfd, u_char *ptr, int len)
{
    int n;

    n = read (sockfd, ptr, len);

    if (n < 0) {

	switch (errno) {
        case EWOULDBLOCK:
#if     defined(EAGAIN) && EAGAIN != EWOULDBLOCK
        case EAGAIN:
#endif  /* EAGAIN */
        case EINTR:
        case ENETUNREACH:
        case EHOSTUNREACH:
	    trace (TR_INFO, peer->trace, 
		   "READ FAILED (%s) -- OKAY TO IGNORE\n", strerror (errno));
	    return (0);
	default:
	    trace (TR_WARN, peer->trace, "READ FAILED (%s)\n", 
		   strerror (errno));
	    return (-1);
	}
    }
    else if (n == 0) {
	trace (TR_WARN, peer->trace, "READ FAILED EOF???\n");
	return (-1);
    }

    trace (TR_PACKET, peer->trace, "READ (%d bytes)\n", n);
    return (n);
}


/*
 * reads data on socket as much as it can
 * returns the length or -1 on error
 * peer->start_ptr will have the pointer
 * peer->read_ptr keeps the end of data
 */
static int
bgp_fill_packet (bgp_peer_t *peer)
{
    int len, n;

    assert (peer);
    assert (peer->read_ptr >= peer->buffer);
    assert (peer->read_ptr <= peer->buffer + sizeof (peer->buffer));
    assert (peer->start_ptr >= peer->buffer);
    assert (peer->start_ptr <= peer->buffer + sizeof (peer->buffer));

    if ((len = peer->read_ptr - peer->start_ptr) == 0) {
	/* reset the pointers */
	peer->start_ptr = peer->buffer;
	peer->read_ptr = peer->buffer;
    }

    if (peer->buffer + sizeof (peer->buffer) - peer->read_ptr 
		< BGPMAXPACKETSIZE) {
	/* need to move them to the start to get more */
	memcpy (peer->buffer, peer->start_ptr, len);
	peer->start_ptr = peer->buffer;
	peer->read_ptr = peer->buffer + len;
    }

    if ((n = bgp_read (peer, peer->sockfd, peer->read_ptr, BGPMAXPACKETSIZE)) 
		< 0) {
	/* will not send notification for this kind of error */
	bgp_sm_process_event (peer, BGPEVENT_ERROR);
	return (-1);
    }
    else if (n == 0) {
	return (0);
    }

    peer->read_ptr += n;
    assert (peer->read_ptr <= peer->buffer + sizeof (peer->buffer));

    return (len + n);
}


/* 
 * Masaki's version is get_packet ()
 * fills in buffer and returns 1 if full one has arrived
 *                     returns 0 if not, -1 on error
 */
static int
get_packet (bgp_peer_t *peer)
{
    int pdu_len, len;
    u_char *cp;

    peer->packet = NULL;

    /* need to be filled at least a header in buffer */
    /* check if the requested length of data already in buffer */
    if ((len = peer->read_ptr - peer->start_ptr) < BGP_HEADER_LEN) {
	return (0);
    }

    cp = peer->start_ptr;

    BGP_GET_HDRLEN (pdu_len, cp);
    if (pdu_len < BGP_HEADER_LEN || pdu_len > BGPMAXPACKETSIZE) {
        peer->start_ptr = peer->read_ptr; /* eat up the input */
        bgp_send_notify_word (peer, BGP_ERR_HEADER, BGP_ERRHDR_LENGTH, pdu_len);
	bgp_sm_process_event (peer, BGPEVENT_ERROR);
	return (-1);
    }

    /* see if the total length packet in buffer */
    /* check if the requested length of data already in buffer */
    if (len < pdu_len) {
	return (0);
    }

    peer->packet = peer->start_ptr;
    peer->start_ptr += pdu_len;
    return (1);
}


/* 
 * munge bgp packet
 * Process notifications here -- deal with errors immediately
 */
static int 
process_pdu (bgp_peer_t * peer)
{
    int type;
    int length;
    int error = 0;
    u_char *cp;

    assert (peer);
    assert (peer->packet);
    cp = peer->packet;
    BGP_GET_HDRTYPE (type, cp);
    BGP_GET_HDRLEN (length, cp);

    if (type >= BGP_OPEN && type < BGP_PACKET_MAX) {
        trace (TR_PACKET, peer->trace, "recv %s (%d bytes)\n",
	       sbgp_pdus[type], length);
    }

    switch (type) {
        case BGP_OPEN:
	    if (length < BGP_HEADER_LEN+BGP_OPEN_MIN_LEN) {
		error = 1;
		break;
	    }
	    bgp_sm_process_event (peer, BGPEVENT_RECVOPEN);
	    break;
	case BGP_UPDATE:
	    if (length < BGP_HEADER_LEN+BGP_UPDATE_MIN_LEN) {
		error = 1;
		break;
	    }
	    bgp_sm_process_event (peer, BGPEVENT_RECVUPDATE);
	    break;
	case BGP_NOTIFY:
	    if (length < BGP_HEADER_LEN+BGP_NOTIFY_MIN_LEN) {
		error = 1;
		break;
	    }
	    bgp_process_notify (peer);
	    bgp_sm_process_event (peer, BGPEVENT_RECVNOTIFY);
	    break;
	case BGP_KEEPALIVE:
	    if (length != BGP_HEADER_LEN) {
		error = 1;
		break;
	    }
	    bgp_sm_process_event (peer, BGPEVENT_RECVKEEPALIVE);
	    break;
	default:
            trace (TR_ERROR, peer->trace, 
		"recv unknown message type %d (%d bytes)\n", type, length);
	    bgp_send_notify_byte (peer, BGP_ERR_HEADER, BGP_ERRHDR_TYPE, type);
	    bgp_sm_process_event (peer, BGPEVENT_ERROR);
	    return (-1);
    }
    if (error > 0) {
        trace (TR_ERROR, peer->trace, 
	    "recv %s with bad length %d\n", sbgp_pdus[type], length);
	bgp_send_notify_word (peer, BGP_ERR_HEADER, BGP_ERRHDR_LENGTH, length);
	bgp_sm_process_event (peer, BGPEVENT_ERROR);
	return (-1);
    }
    return (1);
}


/*
 * called after mrt_select notices change on descriptor. 
 * Gets packet from peer and starts processing
 */
int
bgp_get_pdu (bgp_peer_t * peer)
{
    int ret;

    /* I know that the return value will not be used, but leave as it was */

    if ((ret = bgp_fill_packet (peer)) < 0) {
	/* sockfd should be closed and deleted */
        assert (peer->sockfd < 0);
	return (-1);
    }
    else if (ret == 0) {
        assert (peer->sockfd >= 0);
	select_enable_fd (peer->sockfd);
	return (ret);
    }

    for (;;) {

        if ((ret = get_packet (peer)) < 0) {
	    /* sockfd should be closed and deleted */
            assert (peer->sockfd < 0);
	    break;
        }
        else if (ret == 0) {
	    /* notification may close sockfd */
	    break;
        }
   
        if ((ret = process_pdu (peer)) < 0) {
	    /* sockfd should be closed and deleted */
            assert (peer->sockfd < 0);
	    break;
        }
    }

    if (peer->sockfd >= 0)
        select_enable_fd (peer->sockfd);
    return (1);
}


/* bgp_process_update
 * munge update pdu info into withdrawn routes,  announced routes and 
 * aspath
 */
int 
bgp_process_update (bgp_peer_t * peer)
{
    int type, length;
    u_char *cp;

    assert (peer);
    assert (peer->packet);
    cp = peer->packet;
    BGP_GET_HDRTYPE (type, cp);
    BGP_GET_HDRLEN (length, cp);
    assert (type == BGP_UPDATE);
    assert (length >= BGP_HEADER_LEN+BGP_UPDATE_MIN_LEN && 
	    length <= BGPMAXPACKETSIZE);
    BGP_SKIP_HEADER (cp);
    length -= BGP_HEADER_LEN;

    if (BGP->update_call_fn)
	BGP->update_call_fn (peer, cp, length);

    return (1);
}


/* bgp_process_open
 * cp will be used to process another connection in case of racing
 */
int 
bgp_process_open (bgp_peer_t * peer)
{
    int version;
    int as;
    u_long holdtime;
    int authcode;
    u_long id;
    char tmpx[MAXLINE];
    u_char * cp = peer->packet;

    BGP_SKIP_HEADER (cp);
    BGP_GET_OPEN_V4 (version, as, holdtime, id, authcode, cp);

    if (version != DEFAULT_BGP_VERSION) {
	trace (TR_ERROR, peer->trace,
	       "running BGP 3 or earlier! (version = %d)\n", version);
	bgp_send_notify_word (peer, BGP_ERR_OPEN, BGP_ERROPN_VERSION,
			      DEFAULT_BGP_VERSION);
	return (-1);
    }

    trace (TR_PACKET, peer->trace,
	   "recv OPEN AS %d, HOLDTIME %d, ID %s, AUTH %d\n",
	   as, holdtime, inet_ntop (AF_INET, &id, tmpx, sizeof (tmpx)), 
	   authcode);

    if (id == 0L || id == ~0L) {
        trace (TR_ERROR, peer->trace, "invalid router id: %s\n",
	       inet_ntop (AF_INET, &id, tmpx, sizeof (tmpx)));
	bgp_send_notification (peer, BGP_ERR_OPEN, BGP_ERROPN_BGPID);
	return (-1);
    }

    if (authcode != 0) {
        trace (TR_ERROR, peer->trace, "unknown athentication code: %d\n",
	       authcode);
	bgp_send_notification (peer, BGP_ERR_OPEN, BGP_ERROPN_AUTHCODE);
	return (-1);
    }

    /* set as if default (peer_as = 0) and check if valid AS */
    /* XXX bad way to change the gateway structure */
    if (peer->gateway->AS == 0)
	peer->gateway->AS = as;
    else if (peer->gateway->AS != as) {
	bgp_send_notification (peer, BGP_ERR_OPEN, BGP_ERROPN_AS);
	return (-1);
    }

    if (holdtime < BGP_MIN_HOLDTIME) {
	bgp_send_notification (peer, BGP_ERR_OPEN, BGP_ERROPN_BADHOLDTIME);
	return (-1);
    }

    peer->gateway->routerid = id;
    if (peer->HoldTime_Interval < 0) { /* not enforced by user */

	if (BGP->Default_HoldTime_Interval > holdtime) {
	    assert (peer->timer_HoldTime);
	    Timer_Set_Time (peer->timer_HoldTime, holdtime);
            trace (TR_INFO, peer->trace, "Set holdtime to %d\n", holdtime);

	    if (peer->HoldTime_Interval < 0) { /* not enforced by user */
	        int one_third = holdtime / 3;
	        if (one_third < 1)
		    one_third = 1;
	        assert (peer->timer_KeepAlive);
	        Timer_Set_Time (peer->timer_KeepAlive, one_third);
                trace (TR_INFO, peer->trace, "Set keepalive to %d\n", 
		       one_third);
	    }
	}
    }

    if (holdtime > 0) {
	Timer_Turn_ON (peer->timer_KeepAlive);
	Timer_Reset_Time (peer->timer_HoldTime);
    }
    else {
	/* keepalive remains off */
	Timer_Turn_OFF (peer->timer_HoldTime);
    }

    return (1);
}


/* bgp_process_notify
 */
int 
bgp_process_notify (bgp_peer_t * peer)
{
    u_char *cp = peer->packet;
    u_char code, subcode;
    int length, datalen, data;

    BGP_GET_HDRLEN (length, cp);
    BGP_SKIP_HEADER (cp);

    BGP_GET_NOTIFY(code, subcode, cp);
    datalen = length - (BGP_HEADER_LEN+BGP_NOTIFY_MIN_LEN);

    peer->code = code;
    peer->subcode = subcode;

    if (datalen == 1) {
	BGP_GET_BYTE (data, cp);
        trace (TR_ERROR, peer->trace, 
	       "recv NOTIFICATION %d/%d: %s with byte data %d\n",
	       code, subcode, bgp_notify_to_string (code, subcode), data);
    }
    else if (datalen == 2) {
	BGP_GET_SHORT (data, cp);
        trace (TR_ERROR, peer->trace, 
	       "recv NOTIFICATION %d/%d: %s with word data %d\n",
	       code, subcode, bgp_notify_to_string (code, subcode), data);
    }
    else if (datalen != 0) {
        trace (TR_ERROR, peer->trace, 
	       "recv NOTIFICATION %d/%d: %s with bad data len %d\n",
	       code, subcode, bgp_notify_to_string (code, subcode), datalen);
	return (-1);
    }
    else {
        trace (TR_ERROR, peer->trace, "recv NOTIFICATION %d/%d: %s\n",
	       code, subcode, bgp_notify_to_string (code, subcode));
    }

    return (1);
}


static int 
bgp_send_packet (int sockfd, int type, int len, u_char *data)
{
    u_char buffer[BGPMAXPACKETSIZE], *cp = buffer;

    assert (len >= 0 && len <= BGPMAXPACKETSIZE - BGP_HEADER_LEN);

    /* set auth field */
    memset (buffer, 0xff, BGP_HEADER_MARKER_LEN);

    BGP_PUT_HDRLEN (BGP_HEADER_LEN + len, cp);
    BGP_PUT_HDRTYPE (type, cp);

    if (len > 0) {
        BGP_SKIP_HEADER (cp);
	memcpy (cp, data, len);
    }

    if (write (sockfd, buffer, BGP_HEADER_LEN + len) != BGP_HEADER_LEN + len)
	return (-1);

    return (0);
}


/* bgp_send_open
 * send an open pdu
 */
int 
bgp_send_open (bgp_peer_t * peer)
{
    u_char buffer[BGPMAXPACKETSIZE], *cp = buffer;
    u_char tmpx[MAXLINE];
    u_char authcode = 0;
    int holdtime = BGP->Default_HoldTime_Interval;

    if (peer->HoldTime_Interval >= 0) /* specified by user */
	holdtime = peer->HoldTime_Interval;

    BGP_PUT_OPEN (DEFAULT_BGP_VERSION, BGP->my_as, holdtime,
		  BGP->my_id, authcode, cp);

    if (bgp_send_packet (peer->sockfd, BGP_OPEN, cp - buffer, buffer) < 0) {
	trace (TR_ERROR, peer->trace, "send OPEN failed\n");
	bgp_sm_process_event (peer, BGPEVENT_ERROR);
	return (-1);
    }

    trace (TR_PACKET, peer->trace,
	"send OPEN with own AS %d, HOLDTIME %d, ID %s, AUTH %d\n",
	   BGP->my_as, holdtime,
	   inet_ntop (AF_INET, &BGP->my_id, tmpx, sizeof (tmpx)), authcode);

    /* peer->HoldTime_Interval = holdtime; */
    Timer_Turn_ON (peer->timer_HoldTime);
    return (1);
}



/* bgp_send_keepalive
 */
int 
bgp_send_keepalive (bgp_peer_t * peer)
{
    if (bgp_send_packet (peer->sockfd, BGP_KEEPALIVE, 0, NULL) < 0) {
	trace (TR_ERROR, peer->trace, "send KEEPALIVE failed\n");
	bgp_sm_process_event (peer, BGPEVENT_ERROR);
	return (-1);
    }
    trace (TR_PACKET, peer->trace, "send KEEPALIVE\n");
    return (1);
}


/* bgp_send_notification
 * Send notification error message to sockfd
 * prefix and port are informational
 */
int 
bgp_send_notification2 (int sockfd, prefix_t * prefix, int port,
		       int code, int subcode)
{
    u_char note[2], *cp = note;

    BGP_PUT_BYTE (code, cp);
    BGP_PUT_BYTE (subcode, cp);

    if (bgp_send_packet (sockfd, BGP_NOTIFY, cp - note, note) < 0) {
	trace (TR_ERROR, BGP->trace, "with %s (port %d) send NOTIFY failed\n", 
		   prefix_toa (prefix), port);
	return (-1);
    }

    trace (TR_WARN, BGP->trace, 
	   "with %s (port %d) send NOTIFICATION %d/%d: %s\n",
	   prefix_toa (prefix), port, code, subcode, 
	   bgp_notify_to_string (code, subcode));
    return (1);
}


static int 
bgp_send_notify (bgp_peer_t * peer, int code, int subcode,
		      u_char *data, int len)
{
    u_char note[BGPMAXPACKETSIZE], *cp = note;

    assert (peer);

    BGP_PUT_NOTIFY (code, subcode, len, data, cp);

    if (bgp_send_packet (peer->sockfd, BGP_NOTIFY, cp - note, note) < 0) {
	trace (TR_ERROR, peer->trace, "send NOTIFY failed\n");
        /* bgp_sm_process_event (peer, BGPEVENT_ERROR); */
	return (-1);
    }

    trace (TR_WARN, peer->trace, 
	   "send NOTIFICATION %d/%d: %s with data %d bytes\n", code, subcode, 
	   bgp_notify_to_string (code, subcode), len);

    return (1);
}


/* bgp_send_notification for the peer
 */
int 
bgp_send_notification (bgp_peer_t * peer, int code, int subcode)
{
    return (bgp_send_notify (peer, code, subcode, NULL, 0));
}


int 
bgp_send_notify_byte (bgp_peer_t * peer, int code, int subcode, int opt)
{
    u_char data[1];

    data[0] = opt;
    return (bgp_send_notify (peer, code, subcode, data, sizeof data));
}


int 
bgp_send_notify_word (bgp_peer_t * peer, int code, int subcode, int opt)
{
    u_char data[2];

    data[0] = (opt >> 8) & 0xff;
    data[1] = opt & 0xff;
    return (bgp_send_notify (peer, code, subcode, data, sizeof data));
}


/*-----------------------------------------------------------
 *  Name: 	bgp_send_update
 *  Created:	Tue Jan 24 09:59:38 1995
 *  Author: 	Craig Labovitz   <labovit@snoopy.merit.net>
 *  DESCR:  	
 */

int 
bgp_send_update (bgp_peer_t * peer, int len, u_char * data)
{
    if (bgp_send_packet (peer->sockfd, BGP_UPDATE, len, data) < 0) {
	trace (TR_ERROR, peer->trace, "send UPDATE failed\n");
	bgp_sm_process_event (peer, BGPEVENT_ERROR);
	return (-1);
    }
    trace (TR_PACKET, peer->trace, "send UPDATE (%d bytes)\n",
	   BGP_HEADER_LEN + len);

    Delete (data);
    return (1);
}
